import logging
from datetime import datetime
from time import sleep

from slackminion.dispatcher import MessageDispatcher
from slackminion.slack import SlackEvent, SlackUser, SlackRoomIMBase, ThreadedSlackClient
from slackminion.exceptions import NotSetupError
from slackminion.plugin import PluginManager
from slackminion.webserver import Webserver

from six import string_types


def eventhandler(*args, **kwargs):
    """
    Decorator.  Marks a function as a receiver for the specified slack event(s).

    * events - String or list of events to handle
    """

    def wrapper(func):
        if isinstance(kwargs['events'], string_types):
            kwargs['events'] = [kwargs['events']]
        func.is_eventhandler = True
        func.events = kwargs['events']
        return func

    return wrapper


class Bot(object):
    def __init__(self, config, test_mode=False):
        self.always_send_dm = []
        self.config = config
        self.dispatcher = MessageDispatcher()
        self.event_handlers = {}
        self.is_setup = False
        self.log = logging.getLogger(type(self).__name__)
        self.plugins = PluginManager(self, test_mode)
        self.runnable = True
        self.sc = None
        self.webserver = None
        self.test_mode = test_mode
        self.reconnect_needed = True
        self.bot_start_time = None
        self.timers = []

        if self.test_mode:
            self.metrics = {
                'startup_time': 0
            }

        from slackminion.plugins.core import version
        self.version = version
        try:
            from slackminion.plugins.core import commit
            self.commit = commit
        except ImportError:
            self.commit = "HEAD"

    def start(self):
        """Initializes the bot, plugins, and everything."""
        self.bot_start_time = datetime.now()
        self.webserver = Webserver(self.config['webserver']['host'], self.config['webserver']['port'])
        self.plugins.load()
        self.plugins.load_state()
        self._find_event_handlers()
        self.sc = ThreadedSlackClient(self.config['slack_token'])

        self.always_send_dm = ['_unauthorized_']
        if 'always_send_dm' in self.config:
            self.always_send_dm.extend(map(lambda x: '!' + x, self.config['always_send_dm']))

        # Rocket is very noisy at debug
        logging.getLogger('Rocket.Errors.ThreadPool').setLevel(logging.INFO)

        self.is_setup = True
        if self.test_mode:
            self.metrics['startup_time'] = (datetime.now() - self.bot_start_time).total_seconds() * 1000.0

    def _find_event_handlers(self):
        for name in dir(self):
            method = getattr(self, name)
            if callable(method) and hasattr(method, 'is_eventhandler'):
                for event in method.events:
                    self.event_handlers[event] = method

    def run(self, start=True):
        """
        Connects to slack and enters the main loop.

        * start - If True, rtm.start API is used. Else rtm.connect API is used

        For more info, refer to
        https://python-slackclient.readthedocs.io/en/latest/real_time_messaging.html#rtm-start-vs-rtm-connect
        """
        # Fail out if setup wasn't run
        if not self.is_setup:
            raise NotSetupError

        # Start the web server
        self.webserver.start()

        first_connect = True

        try:
            while self.runnable:
                if self.reconnect_needed:
                    if not self.sc.rtm_connect(with_team_state=start):
                        return False
                    self.reconnect_needed = False
                    if first_connect:
                        first_connect = False
                        self.plugins.connect()

                # Get all waiting events - this always returns a list
                try:
                    events = self.sc.rtm_read()
                except AttributeError:
                    self.log.exception('Something has failed in the slack rtm library.  This is fatal.')
                    self.runnable = False
                    events = []
                except:
                    self.log.exception('Unhandled exception in rtm_read()')
                    self.reconnect_needed = True
                    events = []
                for e in events:
                    try:
                        self._handle_event(e)
                    except KeyboardInterrupt:
                        # Gracefully shutdown
                        self.runnable = False
                    except:
                        self.log.exception('Unhandled exception in event handler')
                sleep(0.1)
        except KeyboardInterrupt:
            # On ctrl-c, just exit
            pass
        except:
            self.log.exception('Unhandled exception')

    def stop(self):
        """Does cleanup of bot and plugins."""
        # cleanup any running timer threads so bot doesn't hang on shutdown
        for t in self.timers:
            t.cancel()
        if self.webserver is not None:
            self.webserver.stop()
        if not self.test_mode:
            self.plugins.save_state()
        self.plugins.unload_all()

    def send_message(self, channel, text, thread=None, reply_broadcast=None):
        """
        Sends a message to the specified channel

        * channel - The channel to send to.  This can be a SlackChannel object, a channel id, or a channel name
        (without the #)
        * text - String to send
        * thread - reply to the thread. See https://api.slack.com/docs/message-threading#threads_party
        * reply_broadcast - Set to true to indicate your reply is germane to all members of a channel
        """
        # This doesn't want the # in the channel name
        if isinstance(channel, SlackRoomIMBase):
            channel = channel.id
        self.log.debug("Trying to send to %s: %s", channel, text)
        self.sc.rtm_send_message(channel, text, thread=thread, reply_broadcast=reply_broadcast)

    def send_im(self, user, text):
        """
        Sends a message to a user as an IM

        * user - The user to send to.  This can be a SlackUser object, a user id, or the username (without the @)
        * text - String to send
        """
        if isinstance(user, SlackUser):
            user = user.id
            channelid = self._find_im_channel(user)
        else:
            channelid = user.id
        self.send_message(channelid, text)

    def _find_im_channel(self, user):
        resp = self.sc.api_call('im.list')
        channels = filter(lambda x: x['user'] == user, resp['ims'])
        if len(channels) > 0:
            return channels[0]['id']
        resp = self.sc.api_call('im.open', user=user)
        return resp['channel']['id']

    def _load_user_rights(self, user):
        if user is not None:
            if 'bot_admins' in self.config:
                if user.username in self.config['bot_admins']:
                    user.is_admin = True

    def _handle_event(self, event):
        if 'type' not in event:
            # This is likely a notification that the bot was mentioned
            self.log.debug("Received odd event: %s", event)
            return
        e = SlackEvent(sc=self.sc, **event)
        self.log.debug("Received event type: %s", e.type)

        if e.user is not None:
            if hasattr(self, 'user_manager'):
                if not isinstance(e.user, SlackUser):
                    self.log.debug("User is not SlackUser: %s", e.user)
                else:
                    user = self.user_manager.get(e.user.id)
                    if user is None:
                        user = self.user_manager.set(e.user)
                    e.user = user

        if e.type in self.event_handlers:
            self.event_handlers[e.type](e)

    @eventhandler(events='message')
    def _event_message(self, msg):
        self.log.debug("Message.message: %s: %s: %s", msg.channel, msg.user, msg.__dict__)

        # The user manager should load rights when a user is added
        if not hasattr(self, 'user_manager'):
            self._load_user_rights(msg.user)
        try:
            cmd, output, cmd_options = self.dispatcher.push(msg)
        except:
            self.log.exception('Unhandled exception')
            return
        self.log.debug("Output from dispatcher: %s", output)
        if output:
            if cmd_options.get('reply_in_thread'):
                if hasattr(msg, 'thread_ts'):
                    thread_ts = msg.thread_ts
                else:
                    thread_ts = msg.ts
            else:
                thread_ts = None
            if cmd in self.always_send_dm or cmd_options.get('always_send_dm'):
                self.send_im(msg.user, output)
            else:
                self.send_message(msg.channel, output, thread=thread_ts, reply_broadcast=cmd_options.get('reply_broadcast'))

    @eventhandler(events='error')
    def _event_error(self, msg):
        self.log.error("Received an error response from Slack: %s", msg.__dict__)

    @eventhandler(events='team_migration_started')
    def _event_team_migration_started(self, msg):
        self.log.warning("Slack has initiated a team migration to a new server.  Attempting to reconnect...")
        self.reconnect_needed = True



from .base import SlackRoomIMBase


class SlackIM(SlackRoomIMBase):
    """Represents an IM 'room'."""
    def __init__(self, *args, **kwargs):
        super(SlackIM, self).__init__(*args, **kwargs)
        self.is_im = True

    @property
    def channel(self):
        self.logger.warning('Use of channel is deprecated, use id instead')
        return self.name

    @property
    def name(self):
        return self.id

    def __str__(self):
        return '<#%s|%s>' % (self.id, self.id)

    def __repr__(self):
        return self.id



from slackminion.bot import Bot
from slackminion.webserver import Webserver


class DummyBot(Bot):
    def __init__(self, *args, **kwargs):
        super(DummyBot, self).__init__(None, *args, **kwargs)
        setattr(self, 'start', lambda: None)
        setattr(self, 'send_message', lambda x, y, z, a: None)
        self.webserver = Webserver('127.0.0.1', '9999')



from datetime import datetime
from flask import render_template
from functools import wraps
from operator import itemgetter

from slackminion.plugin import cmd, webhook
from slackminion.plugin.base import BasePlugin
from slackminion.slack import SlackRoom

from . import version
try:
    from . import commit
except ImportError:
    commit = 'HEAD'


def channel_wrapper(f):
    @wraps(f)
    def wrapper(self, msg, args):
        channel = self._get_channel_from_msg_or_args(msg, args)
        if channel is not None:
            return f(self, channel)
        return None
    return wrapper


class Core(BasePlugin):

    @cmd()
    def help(self, msg, args):
        """Displays help for each command"""
        output = []
        if len(args) == 0:
            commands = sorted(self._bot.dispatcher.commands.items(), key=itemgetter(0))
            commands = filter(lambda x: x[1].is_subcmd is False, commands)
            # Filter commands if auth is enabled, hide_admin_commands is enabled, and user is not admin
            if self._should_filter_help_commands(msg.user):
                commands = filter(lambda x: x[1].admin_only is False, commands)
            for name, cmd in commands:
                output.append(self._get_short_help_for_command(name))
        else:
            name = '!' + args[0]
            output = [self._get_help_for_command(name)]
        return '\n'.join(output)

    def _should_filter_help_commands(self, user):
        return hasattr(self._bot.dispatcher, 'auth_manager') \
            and 'hide_admin_commands' in self._bot.config \
            and self._bot.config['hide_admin_commands'] is True \
            and not getattr(user, 'is_admin', False)

    def _get_help_for_command(self, name):
        if name not in self._bot.dispatcher.commands:
            return 'No such command: %s' % name
        return self._bot.dispatcher.commands[name].formatted_help

    def _get_short_help_for_command(self, name):
        helpstr = self._bot.dispatcher.commands[name].short_help
        return "*{name}*: {help}".format(name=name, help=helpstr)

    @cmd(admin_only=True)
    def save(self, msg, args):
        """Causes the bot to write its current state to backend."""
        self.send_message(msg.channel, "Saving current state...")
        self._bot.plugins.save_state()
        self.send_message(msg.channel, "Done.")

    @cmd(admin_only=True)
    def shutdown(self, msg, args):
        """Causes the bot to gracefully shutdown."""
        self.log.info("Received shutdown from %s", msg.user.username)
        self._bot.runnable = False
        return "Shutting down..."

    @cmd()
    def whoami(self, msg, args):
        """Prints information about the user and bot version."""
        output = ["Hello %s" % msg.user]
        if hasattr(self._bot.dispatcher, 'auth_manager') and msg.user.is_admin is True:
            output.append("You are a *bot admin*.")
        output.append("Bot version: %s-%s" % (self._bot.version, self._bot.commit))
        return '\n'.join(output)

    @cmd()
    @channel_wrapper
    def sleep(self, channel):
        """Causes the bot to ignore all messages from the channel.

        Usage:
        !sleep [channel name] - ignore the specified channel (or current if none specified)
        """
        self.log.info('Sleeping in %s', channel)
        self._bot.dispatcher.ignore(channel)
        self.send_message(channel, 'Good night')

    @cmd(admin_only=True, while_ignored=True)
    @channel_wrapper
    def wake(self, channel):
        """Causes the bot to resume operation in the channel.

        Usage:
        !wake [channel name] - unignore the specified channel (or current if none specified)
        """
        self.log.info('Waking up in %s', channel)
        self._bot.dispatcher.unignore(channel)
        self.send_message(channel, 'Hello, how may I be of service?')

    @webhook('/status', method='GET')
    def bot_status(self):
        # TODO: Plugins should provide a get_status() or similar that
        # outputs html/dict+template name.  This command should read
        # from that.  The below should be provided by core.
        plugins = [{
            'name': type(x).__name__,
            'version': x._version,
            'commit': x._commit,
        } for x in self._bot.plugins.plugins]

        uptime = datetime.now() - self._bot.bot_start_time
        partial_day = uptime.seconds
        u_hours = partial_day / 3600

        partial_day %= 3600
        u_minutes = partial_day / 60
        u_seconds = partial_day % 60
        context = {
            'bot_name': self._bot.sc.server.username,
            'version': self._bot.version,
            'commit': self._bot.commit,
            'plugins': plugins,
            'uptime': {
                'days': uptime.days,
                'hours': u_hours,
                'minutes': u_minutes,
                'seconds': u_seconds,
            },
        }
        return render_template('status.html', **context)

    def _get_channel_from_msg_or_args(self, msg, args):
        channel = None
        if len(args) == 0:
            if isinstance(msg.channel, SlackRoom):
                channel = msg.channel
        else:
            channel = self.get_channel(args[0])
        return channel



class DuplicateCommandError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Command already defined: %s" % self.name


class DuplicatePluginError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Plugin already defined: %s" % self.name


class NotSetupError(Exception):
    def __str__(self):
        return "Bot not setup.  Please run start() before run()."



import logging
import os
import requests

from flask import Flask, request
from threading import Thread
from uuid import uuid4 as uuid


class Webserver(object):
    def __init__(self, host, port):
        self.log = logging.getLogger(type(self).__name__)
        template_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')
        self.log.debug("Using template path: %s", template_path)
        self.app = Flask(__name__, template_folder=template_path)
        self.host = host
        self.port = port
        self.running = False
        self.shutdown_code = uuid().hex
        self.thread = None

    def start(self):
        self.app.debug = False
        webserver_args = {
            'host': self.host,
            'port': self.port,
            'threaded': True,
        }
        self.thread = Thread(target=self.app.run, kwargs=webserver_args)
        self.thread.start()
        self.running = True

        @self.app.route('/_/shutdown', methods=['POST'])
        def shutdown():
            code = request.form.get('code')
            if code == self.shutdown_code:
                self.log.warn("Shutdown code received.  Stopping webserver...")
                request.environ.get('werkzeug.server.shutdown')()
                self.running = False
            return ''

        @self.app.template_filter()
        def pluralize(val, str, suffix='s'):
            if val != 1:
                str += suffix
            return '%d %s' % (val, str)

    def stop(self):
        if self.running:
            requests.post('http://localhost:%d/_/shutdown' % self.port, data={'code': self.shutdown_code})
            self.thread.join()



import textwrap


def format_docstring(docstring):
    """
    Uses textwrap to auto-dedent a docstring (removes leading spaces)
    Returns the docstring in a pre-formatted block with required characters escaped.
    https://api.slack.com/docs/message-formatting
    :param docstring: str
    :return: str
    """
    if not docstring:
        return ''
    formatted_text = '```{}```'.format(
        textwrap.dedent(docstring.replace('&', '&amp;')
                        .replace('<', '&lt;')
                        .replace('>', '&gt;')
                        )
    )
    return formatted_text



from .attributes import SlackRoomAttribute, SlackRoomTopic
from .base import SlackRoomIMBase
from ..user import SlackUser


class SlackRoom(SlackRoomIMBase):
    """Base class for channels and groups.  Provides handling for loading/setting extra attributes"""
    BASE_ATTRIBUTES = [
        'created',
        'creator',
        'is_archived',
        'members',
        'name',
        'purpose',
        ('topic', SlackRoomTopic)
    ]

    def __init__(self, *args, **kwargs):
        super(SlackRoom, self).__init__(*args, **kwargs)

        # Extra information (lazy loaded)
        self._add_extra_attributes()

        if self._sc is not None:
            resp = self._sc.server.channels.find(self.id)
            if resp is not None:
                self.name = resp.name

        for k, v in kwargs.items():
            if k in self.BASE_ATTRIBUTES + self.EXTRA_ATTRIBUTES:
                setattr(self, k, v)

    def set_topic(self, topic):
        self._sc.api_call(self.API_PREFIX + '.setTopic', channel=self.id, topic=topic)

    def _add_extra_attributes(self):
        for attribute in self.BASE_ATTRIBUTES + self.EXTRA_ATTRIBUTES:
            attribute_class = SlackRoomAttribute
            if isinstance(attribute, tuple):
                attribute, attribute_class = attribute
            setattr(SlackChannel, attribute, attribute_class())

    def _load_extra_attributes(self):
        resp = self._sc.api_call(self.API_PREFIX + '.info', channel=self.id)
        for k, v in resp[self.ATTRIBUTE_KEY].items():
            if k == 'creator':
                v = SlackUser(v, sc=self._sc)
            elif k == 'topic':
                v = v['value']
            setattr(self, k, v)

    @staticmethod
    def get_channel(sc, channel_name):
        resp = sc.server.channels.find(channel_name)
        if resp is None:
            return None
        channel_class = SlackChannel
        if resp.id[0] == 'G':
            channel_class = SlackGroup
        channel = channel_class(resp.id, name=resp.name, sc=sc)
        return channel

    @property
    def channel(self):
        self.logger.warning('Use of channel is deprecated, use name instead')
        return self.name

    @property
    def channelid(self):
        self.logger.warning('Use of channelid is deprecated, use id instead')
        return self.id

    def __str__(self):
        return '<#%s|%s>' % (self.id, self.name)

    def __repr__(self):
        return self.id


class SlackChannel(SlackRoom):
    """Represents a Slack channel"""
    API_PREFIX = 'channels'
    ATTRIBUTE_KEY = 'channel'
    EXTRA_ATTRIBUTES = [
        'is_channel',
        'is_general',
    ]


class SlackGroup(SlackChannel):
    """Represents a Slack group (private channel)"""
    API_PREFIX = 'groups'
    ATTRIBUTE_KEY = 'group'
    EXTRA_ATTRIBUTES = [
        'is_group',
        'is_mpim',
    ]



import threading

from slackclient.server import Server
from slackclient.client import SlackClient


class ThreadedSlackClient(SlackClient):
    def __init__(self, *args, **kwargs):
        super(ThreadedSlackClient, self).__init__(*args, **kwargs)
        self.server = ThreadedServer(self.token, False)


class ThreadedServer(Server):
    def __init__(self, *args, **kwargs):
        super(ThreadedServer, self).__init__(*args, **kwargs)
        self.api_call_lock = threading.RLock()

    def api_call(self, *args, **kwargs):
        with self.api_call_lock:
            return super(ThreadedServer, self).api_call(*args, **kwargs)



from slackclient.channel import Channel
from slackclient.user import User
from slackclient.util import SearchList

from slackminion.slack import SlackEvent

test_channel_id = 'C12345678'
test_channel_name = 'testchannel'
test_group_id = 'G12345678'
test_group_name = 'testgroup'
test_user_id = 'U12345678'
test_user_name = 'testuser'
test_user_email = 'root@dev.null'


class DummyServer(object):
    def __init__(self):
        self.channels = SearchList()
        self.users = SearchList()


class DummySlackConnection(object):
    def __init__(self):
        self.server = DummyServer()
        self.server.channels.append(Channel(None, test_channel_name, test_channel_id))
        self.server.channels.append(Channel(None, test_group_name, test_group_id))
        self.server.users.append(User(None, test_user_name, test_user_id, test_user_name, None, test_user_email))
        self.topic = 'Test Topic'

    def api_call(self, name, *args, **kwargs):
        if 'setTopic' in name:
            self.topic = kwargs['topic']
        api_responses = {
            'channels.info': {
                'channel': {
                    'name': test_channel_name,
                    'creator': test_user_id,
                    'topic': {
                        'value': self.topic,
                    },
                },
            },
            'channels.setTopic': {
                'topic': self.topic,
            },
            'groups.info': {
                'group': {
                    'name': test_group_name,
                    'creator': test_user_id,
                    'topic': {
                        'value': self.topic,
                    },
                },
            },
            'groups.setTopic': {
                'topic': self.topic,
            },
            'users.info': {
                'user': {
                    'id': test_user_id,
                    'name': test_user_name,
                },
            }
        }
        return api_responses[name]


def get_test_event():
    return SlackEvent(sc=DummySlackConnection(), user=test_user_id, channel=test_channel_id)



import pytest

from slackminion.plugin import BasePlugin, cmd

from .bot import DummyBot
from .slack import DummySlackConnection

EXPECTED_PLUGIN_METHODS = [
    'on_connect',
    'on_load',
    'on_unload',
]


class DummyPlugin(BasePlugin):
    @cmd(aliases='bca')
    def abc(self, msg, args):
        return 'abcba'

    @cmd(aliases='gfe', reply_in_thread=True)
    def efg(self, msg, args):
        return 'efgfe'

    @cmd(aliases='jih', reply_in_thread=True, reply_broadcast=True)
    def hij(self, msg, args):
        return 'hijih'


class BasicPluginTest(object):
    PLUGIN_CLASS = None
    BASE_METHODS = ['on_load', 'on_connect', 'on_unload']
    ADMIN_COMMANDS = []

    def setup(self):
        self.bot = DummyBot()
        self.bot.sc = DummySlackConnection()
        self.object = self.PLUGIN_CLASS(self.bot)
        self.called = {}

    def teardown(self):
        self.object = None

    def is_called(self, method, test_func, *args, **kwargs):
        from _pytest.monkeypatch import MonkeyPatch

        def called_func(*args, **kwargs):
            self.called[method] = True
        self.called[method] = False
        m = MonkeyPatch()
        m.setattr(method, called_func)
        test_func(*args, **kwargs)
        return self.called[method]

    @pytest.mark.parametrize('method', EXPECTED_PLUGIN_METHODS)
    def test_has_base_method(self, method):
        assert hasattr(self.object, method)

    @pytest.mark.parametrize('method', EXPECTED_PLUGIN_METHODS)
    def test_method_returns_true(self, method):
        f = getattr(self.object, method)
        assert f() is True

    def test_admin_commands(self):
        for c in self.ADMIN_COMMANDS:
            assert getattr(self.object, c).admin_only is True



from setuptools import setup, find_packages
from slackminion.plugins.core import version

setup(
        name='slackminion',
        version=version,
        packages=find_packages(exclude=['test_plugins']),
        url='https://github.com/arcticfoxnv/slackminion',
        license='MIT',
        author='Nick King',
        author_email='',
        description='A python bot framework for slack',
        package_data={'slackminion': ['templates/*']},
        install_requires=[
            'Flask>=0.10.1',
            'itsdangerous>=0.24',
            'Jinja2>=2.6',
            'MarkupSafe>=0.23',
            'PyYAML>=3.10',
            'requests >=2.11, <3.0a0',
            'six >=1.10, <2.0a0',
            'slackclient==1.3.1',
            'websocket-client >=0.35, <0.55.0',
            'Werkzeug>=0.10.4',
        ],
        setup_requires=[
            'pytest-runner'
        ],
        tests_require=[
            'pytest==4.4.1',
            'pytest-cov==2.6.1',
            'codeclimate-test-reporter==0.1.2',
            'coverage==4.5.2',
            'mock==3.0.5',
        ],
        entry_points={
            'console_scripts': [
                'slackminion = slackminion.__main__:main',
            ]
        },
        classifiers=[
            "Programming Language :: Python",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Development Status :: 2 - Pre-Alpha",
            "Topic :: Communications :: Chat",
        ]
)
